﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;

namespace ContinueWithVsAsyncMethods {
	class Program {
		static void Main(string[] args) {
			Console.WriteLine("Main: Thread {0}: Result of Process1 = {1} ...", TID, Process1Async(1).Result); Console.WriteLine("------");
			Console.WriteLine("Main: Thread {0}: Result of Process1 = {1} ...", TID, Process1bAsync(1).Result); Console.WriteLine("------");
			Console.WriteLine();
			Console.WriteLine("Main: Thread {0}: Result of Process2 = {1} ...", TID, Process2Async(1).Result); Console.WriteLine("------");
			Console.WriteLine();
			Console.WriteLine("Main: Thread {0}: Result of Process3 = {1} ...", TID, Process3Async(1).Result);
		}

		static int TID {
			get { return System.Threading.Thread.CurrentThread.ManagedThreadId; }
		}

		static Task<int> DoSomeIOAsync(int param) {
			return Task<int>.Factory.StartNew(() => {
				Console.WriteLine("    Thread {0}: DoSomeIOAsync({1})", TID, param);
				System.Threading.Thread.Sleep(100);
				return param;
			}, TaskCreationOptions.LongRunning);
		}

		static Task<int> Process1Async(int value) {
			Console.WriteLine("Thread {0}: Process1: prepare phase ...", TID);
			return Task<int>.Run(() => {
				Task<int> t1 = DoSomeIOAsync(value);
				Console.WriteLine("Thread {0}: Process1: processing value {1} ...", TID, t1.Result);

				Task<int> t2 = DoSomeIOAsync(value + 1);
				Console.WriteLine("Thread {0}: Process1: processing value {1} ...", TID, t2.Result);
	
				Task<int> t3 = DoSomeIOAsync(value + 2);
				Console.WriteLine("Thread {0}: Process1: processing value {1} ...", TID, t3.Result);

				return t3.Result;
			});
		}

		static Task<int> Process1bAsync(int value) {
			Console.WriteLine("Thread {0}: Process1b: prepare phase ...", TID);
			return Task<int>.Run(() => {
				Task<int> t1 = DoSomeIOAsync(value);
				t1.Wait();
				Console.WriteLine("Thread {0}: Process1b: processing value {1} ...", TID, t1.Result);

				Task<int> t2 = DoSomeIOAsync(value + 1);
				t2.Wait();
				Console.WriteLine("Thread {0}: Process1b: processing value {1} ...", TID, t2.Result);

				Task<int> t3 = DoSomeIOAsync(value + 2);
				t3.Wait();
				Console.WriteLine("Thread {0}: Process1b: processing value {1} ...", TID, t3.Result);

				return t3.Result;
			});
		}

		// Notice that now the continuations run in different threads, thus we are not blocking a single ThreadPool thread as in variants 1a and 1b!
		static Task<int> Process2Async(int value) {
			Console.WriteLine("Thread {0}: Process2: prepare phase ...", TID);
			var tsc = new TaskCompletionSource<int>();
			DoSomeIOAsync(value).ContinueWith((t1) => {
				Console.WriteLine("Thread {0}: Process2: processing value {1} ...", TID, t1.Result);

				DoSomeIOAsync(value + 1).ContinueWith((t2) => {
					Console.WriteLine("Thread {0}: Process2: processing value {1} ...", TID, t2.Result);

					DoSomeIOAsync(value + 2).ContinueWith((t3) => {
						Console.WriteLine("Thread {0}: Process2: processing value {1} ...", TID, t3.Result);
						tsc.SetResult(t3.Result);
					});
				});
			});
			return tsc.Task;
		}

		static async Task<int> Process3Async(int value) {
			Console.WriteLine("Thread {0}: Process3: prepare phase ...", TID);

			int r1 = await DoSomeIOAsync(value);
			Console.WriteLine("Thread {0}: Process3: processing value {1} ...", TID, r1);

			int r2 = await DoSomeIOAsync(value + 1);
			Console.WriteLine("Thread {0}: Process3: processing value {1} ...", TID, r2);

			int r3 = await DoSomeIOAsync(value + 2);
			Console.WriteLine("Thread {0}: Process3: processing value {1} ...", TID, r3);

			return r3;
		}
	}
}
